@@ -1,27 +1,46 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
-/*
- * E-table-column-view.c: A canvas item based view of the ETableColumn.
+/* 
+ * e-table-header-item.c
+ * Copyright 1999, 2000, 2001, Ximian, Inc.
  *
- * Author:
+ * Authors:
+ *   Chris Lahey <clahey@ximian.com>
  *   Miguel de Icaza (miguel@gnu.org)
  *
- * Copyright 1999, 2000, 2001 Ximian, Inc.
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Library General Public
+ * License, version 2, as published by the Free Software Foundation.
+ *
+ * This library is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Library General Public License for more details.
+ *
+ * You should have received a copy of the GNU Library General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
+ * 02111-1307, USA.
  */
+
 #include <config.h>
+
+#include "e-table-header-item.h"
+
+#include <string.h>
 #include <gtk/gtksignal.h>
 #include <gtk/gtkdnd.h>
-#include "gal/util/e-i18n.h"
 #include <libgnomeui/gnome-canvas.h>
 #include <libgnomeui/gnome-canvas-util.h>
 #include <libgnomeui/gnome-canvas-polygon.h>
 #include <libgnomeui/gnome-canvas-rect-ellipse.h>
 #include <gdk-pixbuf/gdk-pixbuf.h>
 #include "gal/widgets/e-cursors.h"
+#include "gal/util/e-i18n.h"
+#include "gal/util/e-util.h"
 #include "gal/util/e-xml-utils.h"
 #include "gal/widgets/e-canvas.h"
 #include "gal/widgets/e-popup-menu.h"
 #include "e-table-header.h"
-#include "e-table-header-item.h"
 #include "e-table-header-utils.h"
 #include "e-table-col-dnd.h"
 #include "e-table-defines.h"
@@ -78,6 +97,7 @@ enum {
 	ARG_TABLE_FONTSET,
 	ARG_SORT_INFO,
 	ARG_TABLE,
+	ARG_TREE,
 };
 
 static void
@@ -152,7 +172,6 @@ ethi_update (GnomeCanvasItem *item, double *affine, ArtSVP *clip_path, int flags
 	if (GNOME_CANVAS_ITEM_CLASS (ethi_parent_class)->update)
 		(*GNOME_CANVAS_ITEM_CLASS (ethi_parent_class)->update)(item, affine, clip_path, flags);
 
-
 	if (ethi->sort_info)
 		ethi->group_indent_width = e_table_sort_info_grouping_get_count(ethi->sort_info) * GROUP_INDENT;
 	else
@@ -191,6 +210,7 @@ ethi_font_set (ETableHeaderItem *ethi, GdkFont *font)
 		gdk_font_unref (ethi->font);
 
 	ethi->font = font;
+	gdk_font_ref (font);
 	
 	ethi->height = e_table_header_item_get_height (ethi);
 	e_canvas_item_request_reflow(GNOME_CANVAS_ITEM(ethi));
@@ -199,13 +219,18 @@ ethi_font_set (ETableHeaderItem *ethi, GdkFont *font)
 static void
 ethi_font_load (ETableHeaderItem *ethi, char *fontname)
 {
-	GdkFont *font;
+	GdkFont *font = NULL;
 
-	font = gdk_fontset_load (fontname);
-	if (font == NULL)
-		font = gdk_font_load ("-adobe-helvetica-medium-r-normal--*-120-*-*-*-*-iso8859-1");
+	if (fontname != NULL)
+		font = gdk_fontset_load (fontname);
+
+	if (font == NULL) {
+		font = GTK_WIDGET (GNOME_CANVAS_ITEM (ethi)->canvas)->style->font;
+		gdk_font_ref (font);
+	}
 	
 	ethi_font_set (ethi, font);
+	gdk_font_unref (font);
 }
 
 static void
@@ -323,6 +348,12 @@ ethi_set_arg (GtkObject *o, GtkArg *arg, guint arg_id)
 		else
 			ethi->table = NULL;
 		break;
+	case ARG_TREE:
+		if (GTK_VALUE_OBJECT(*arg))
+			ethi->tree = E_TREE(GTK_VALUE_OBJECT(*arg));
+		else
+			ethi->tree = NULL;
+		break;
 	}
 	gnome_canvas_item_request_update(item);
 }
@@ -555,7 +586,7 @@ ethi_drag_motion (GtkObject *canvas, GdkDragContext *context,
 		
 		col = ethi_find_col_by_x_nearest (ethi, x);
 
-		if (col == ethi->drag_col || col == ethi->drag_col + 1) {
+		if (ethi->drag_col != -1 && (col == ethi->drag_col || col == ethi->drag_col + 1)) {
 			if (ethi->drag_col != -1)
 				ethi_remove_destroy_marker (ethi);
 			
@@ -605,29 +636,35 @@ ethi_drag_data_received (GtkWidget *canvas,
 			 ETableHeaderItem *ethi)
 {
 	int found = FALSE;
-	int count = e_table_header_count(ethi->eth);
-	int column = atoi(data->data);
-	int drop_col = ethi->drop_col;
+	int count;
+	int column;
+	int drop_col;
 	int i;
-	ethi->drop_col = -1;
 
-	if (column < 0)
-		return;
-	for (i = 0; i < count; i++) {
-		ETableCol *ecol = e_table_header_get_column (ethi->eth, i);
-		if (ecol->col_idx == column) {
-			e_table_header_move(ethi->eth, i, drop_col);
-			found = TRUE;
-			break;
-		}
-	}
-	if (!found) {
-		count = e_table_header_count(ethi->full_header);
-		for (i = 0; i < count; i++) {
-			ETableCol *ecol = e_table_header_get_column (ethi->full_header, i);
-			if (ecol->col_idx == column) {
-				e_table_header_add_column (ethi->eth, ecol, drop_col);
-				break;
+	if (data->data) {
+		count = e_table_header_count(ethi->eth);
+		column = atoi(data->data);
+		drop_col = ethi->drop_col;
+		ethi->drop_col = -1;
+
+		if (column >= 0) {
+			for (i = 0; i < count; i++) {
+				ETableCol *ecol = e_table_header_get_column (ethi->eth, i);
+				if (ecol->col_idx == column) {
+					e_table_header_move(ethi->eth, i, drop_col);
+					found = TRUE;
+					break;
+				}
+			}
+			if (!found) {
+				count = e_table_header_count(ethi->full_header);
+				for (i = 0; i < count; i++) {
+					ETableCol *ecol = e_table_header_get_column (ethi->full_header, i);
+					if (ecol->col_idx == column) {
+						e_table_header_add_column (ethi->eth, ecol, drop_col);
+						break;
+					}
+				}
 			}
 		}
 	}
@@ -699,7 +736,6 @@ ethi_realize (GnomeCanvasItem *item)
 {
 	ETableHeaderItem *ethi = E_TABLE_HEADER_ITEM (item);
 	GdkWindow *window;
-	GdkColor c;
 	GtkTargetEntry  ethi_drop_types [] = {
 		{ TARGET_ETABLE_COL_TYPE, 0, TARGET_ETABLE_COL_HEADER },
 	};
@@ -710,10 +746,6 @@ ethi_realize (GnomeCanvasItem *item)
 
 	window = GTK_WIDGET (item->canvas)->window;
 
-	ethi->gc = gdk_gc_new (window);
-	gnome_canvas_get_color (item->canvas, "black", &c);
-	gdk_gc_set_foreground (ethi->gc, &c);
-
 	if (!ethi->font)
 		ethi_font_set (ethi, GTK_WIDGET (item->canvas)->style->font);
 
@@ -749,8 +781,7 @@ ethi_unrealize (GnomeCanvasItem *item)
 {
 	ETableHeaderItem *ethi = E_TABLE_HEADER_ITEM (item);
 
-	gdk_gc_unref (ethi->gc);
-	ethi->gc = NULL;
+	gdk_font_unref (ethi->font);
 
 	gtk_signal_disconnect (GTK_OBJECT (item->canvas), ethi->drag_motion_id);
 	gtk_signal_disconnect (GTK_OBJECT (item->canvas), ethi->drag_leave_id);
@@ -786,19 +817,19 @@ ethi_draw (GnomeCanvasItem *item, GdkDrawable *drawable, int x, int y, int width
 		for (i = 0; i < length; i++) {
 			ETableSortColumn column = e_table_sort_info_grouping_get_nth(ethi->sort_info, i);
 			g_hash_table_insert (arrows, 
-					     (gpointer) column.column,
-					     (gpointer) (column.ascending ?
-							 E_TABLE_COL_ARROW_UP : 
-							 E_TABLE_COL_ARROW_DOWN));
+					     GINT_TO_POINTER (column.column),
+					     GINT_TO_POINTER (column.ascending ?
+							      E_TABLE_COL_ARROW_UP : 
+							      E_TABLE_COL_ARROW_DOWN));
 		}
 		length = e_table_sort_info_sorting_get_count(ethi->sort_info);
 		for (i = 0; i < length; i++) {
 			ETableSortColumn column = e_table_sort_info_sorting_get_nth(ethi->sort_info, i);
 			g_hash_table_insert (arrows, 
-					     (gpointer) column.column,
-					     (gpointer) (column.ascending ?
-							 E_TABLE_COL_ARROW_UP : 
-							 E_TABLE_COL_ARROW_DOWN));
+					     GINT_TO_POINTER (column.column),
+					     GINT_TO_POINTER (column.ascending ?
+							      E_TABLE_COL_ARROW_UP : 
+							      E_TABLE_COL_ARROW_DOWN));
 		}
 	}
 
@@ -825,12 +856,12 @@ ethi_draw (GnomeCanvasItem *item, GdkDrawable *drawable, int x, int y, int width
 		e_table_header_draw_button (drawable, ecol,
 					    GTK_WIDGET (canvas)->style, ethi->font,
 					    GTK_WIDGET_STATE (canvas),
-					    GTK_WIDGET (canvas), ethi->gc,
+					    GTK_WIDGET (canvas),
 					    x1 - x, -y,
 					    width, height,
 					    x2 - x1, ethi->height,
 					    (ETableColArrow) g_hash_table_lookup (
-						    arrows, (gpointer) ecol->col_idx));
+						    arrows, GINT_TO_POINTER (ecol->col_idx)));
 	}
 
 	g_hash_table_destroy (arrows);
@@ -982,10 +1013,10 @@ ethi_start_drag (ETableHeaderItem *ethi, GdkEvent *event)
 			group_indent ++;
 			g_hash_table_insert (
 				arrows, 
-				(gpointer) column.column,
-				(gpointer) (column.ascending ?
-					    E_TABLE_COL_ARROW_UP : 
-					    E_TABLE_COL_ARROW_DOWN));
+				GINT_TO_POINTER (column.column),
+				GINT_TO_POINTER (column.ascending ?
+						 E_TABLE_COL_ARROW_UP : 
+						 E_TABLE_COL_ARROW_DOWN));
 		}
 		length = e_table_sort_info_sorting_get_count(ethi->sort_info);
 		for (i = 0; i < length; i++) {
@@ -995,10 +1026,10 @@ ethi_start_drag (ETableHeaderItem *ethi, GdkEvent *event)
 
 			g_hash_table_insert (
 				arrows, 
-				(gpointer) column.column,
-				(gpointer) (column.ascending ?
-					    E_TABLE_COL_ARROW_UP : 
-					    E_TABLE_COL_ARROW_DOWN));
+				GINT_TO_POINTER (column.column),
+				GINT_TO_POINTER (column.ascending ?
+						 E_TABLE_COL_ARROW_UP : 
+						 E_TABLE_COL_ARROW_DOWN));
 		}
 	}
 
@@ -1017,12 +1048,12 @@ ethi_start_drag (ETableHeaderItem *ethi, GdkEvent *event)
 		pixmap, ecol,
 		widget->style, ethi->font,
 		GTK_WIDGET_STATE (widget),
-		widget, ethi->gc,
+		widget,
 		0, 0,
 		col_width, ethi->height,
 		col_width, ethi->height,
 		(ETableColArrow) g_hash_table_lookup (
-			arrows, (gpointer) ecol->col_idx));
+			arrows, GINT_TO_POINTER (ecol->col_idx)));
 	gtk_drag_set_icon_pixmap (
 		context,
 		gdk_window_get_colormap (widget->window),
@@ -1084,7 +1115,9 @@ ethi_popup_sort_ascending(GtkWidget *widget, EthiHeaderInfo *info)
 		}
 	}
 	if (!found) {
-		ETableSortColumn column = { model_col, 1 };
+		ETableSortColumn column;
+		column.column = model_col;
+		column.ascending =  1;
 		length = e_table_sort_info_sorting_get_count(ethi->sort_info);
 		if (length == 0)
 			length++;
@@ -1134,7 +1167,9 @@ ethi_popup_sort_descending(GtkWidget *widget, EthiHeaderInfo *info)
 		}
 	}
 	if (!found) {
-		ETableSortColumn column = { model_col, 0 };
+		ETableSortColumn column;
+		column.column = model_col;
+		column.ascending = 0;
 		length = e_table_sort_info_sorting_get_count (ethi->sort_info);
 		if (length == 0)
 			length++;
@@ -1186,6 +1221,7 @@ ethi_popup_field_chooser(GtkWidget *widget, EthiHeaderInfo *info)
 	GtkWidget *etfcd = e_table_field_chooser_dialog_new();
 	gtk_object_set(GTK_OBJECT(etfcd),
 		       "full_header", info->ethi->full_header,
+		       "header", info->ethi->eth,
 		       "dnd_code", info->ethi->dnd_code,
 		       NULL);
 	gtk_widget_show(etfcd);
@@ -1228,7 +1264,10 @@ apply_changes (ETableConfig *config, ETableHeaderItem *ethi)
 {
 	char *state = e_table_state_save_to_string (config->state);
 
-	e_table_set_state (ethi->table, state);
+	if (ethi->table)
+		e_table_set_state (ethi->table, state);
+	if (ethi->tree)
+		e_tree_set_state (ethi->tree, state);
 	g_free (state);
 }
 
@@ -1237,15 +1276,23 @@ ethi_popup_customize_view(GtkWidget *widget, EthiHeaderInfo *info)
 {
 	ETableHeaderItem *ethi = info->ethi;
 	ETableState *state;
+	ETableSpecification *spec;
 
 	if (ethi->config)
 		e_table_config_raise (E_TABLE_CONFIG (ethi->config));
 	else {
-		state = e_table_get_state_object(ethi->table);
+		if (ethi->table) {
+			state = e_table_get_state_object(ethi->table);
+			spec = ethi->table->spec;
+		} else if (ethi->tree) {
+			state = e_tree_get_state_object(ethi->tree);
+			spec = e_tree_get_spec (ethi->tree);
+		} else
+			return;
 
 		ethi->config = e_table_config_new (
-			_("Configuring view: FIXME"),
-			ethi->table->spec, state);
+				_("Configuring view: FIXME"),
+				spec, state);
 		gtk_signal_connect (
 			GTK_OBJECT (ethi->config), "destroy",
 			GTK_SIGNAL_FUNC (config_destroyed), ethi);
@@ -1263,15 +1310,15 @@ static EPopupMenu ethi_context_menu [] = {
 	{ N_("Sort Descending"),           NULL, GTK_SIGNAL_FUNC(ethi_popup_sort_descending), NULL, 2},
 	{ N_("Unsort"),                    NULL, GTK_SIGNAL_FUNC(ethi_popup_unsort),          NULL, 0},
 	{ "",                              NULL, GTK_SIGNAL_FUNC(NULL),                       NULL, 0},
-	{ N_("Group By This Field"),       NULL, GTK_SIGNAL_FUNC(ethi_popup_group_field),     NULL, 0},
-	{ N_("Group By Box"),              NULL, GTK_SIGNAL_FUNC(ethi_popup_group_box),       NULL, 1},
+	{ N_("Group By This Field"),       NULL, GTK_SIGNAL_FUNC(ethi_popup_group_field),     NULL, 16},
+	{ N_("Group By Box"),              NULL, GTK_SIGNAL_FUNC(ethi_popup_group_box),       NULL, 128},
 	{ "",                              NULL, GTK_SIGNAL_FUNC(NULL),                       NULL, 1},
-	{ N_("Remove This Column"),        NULL, GTK_SIGNAL_FUNC(ethi_popup_remove_column),   NULL, 0},
+	{ N_("Remove This Column"),        NULL, GTK_SIGNAL_FUNC(ethi_popup_remove_column),   NULL, 8},
 	{ N_("Add a Column..."),           NULL, GTK_SIGNAL_FUNC(ethi_popup_field_chooser),   NULL, 0},
 	{ "",                              NULL, GTK_SIGNAL_FUNC(NULL),                       NULL, 1},
-	{ N_("Alignment"),                 NULL, GTK_SIGNAL_FUNC(ethi_popup_alignment),       NULL, 1},
+	{ N_("Alignment"),                 NULL, GTK_SIGNAL_FUNC(ethi_popup_alignment),       NULL, 128},
 	{ N_("Best Fit"),                  NULL, GTK_SIGNAL_FUNC(ethi_popup_best_fit),        NULL, 2},
-	{ N_("Format Columns..."),         NULL, GTK_SIGNAL_FUNC(ethi_popup_format_columns),  NULL, 1},
+	{ N_("Format Columns..."),         NULL, GTK_SIGNAL_FUNC(ethi_popup_format_columns),  NULL, 128},
 	{ "",                              NULL, GTK_SIGNAL_FUNC(NULL),                       NULL, 1},
 	{ N_("Customize Current View..."), NULL, GTK_SIGNAL_FUNC(ethi_popup_customize_view),  NULL, 4},
 	{ NULL, NULL, NULL, NULL, 0 }
@@ -1288,7 +1335,10 @@ ethi_header_context_menu (ETableHeaderItem *ethi, GdkEventButton *event)
 	e_popup_menu_run (ethi_context_menu, (GdkEvent *) event,
 			  1 +
 			  (col->sortable ? 0 : 2) +
-			  (ethi->table ? 0 : 4), 0, info);
+			  ((ethi->table || ethi->tree) ? 0 : 4) + 
+			  ((e_table_header_count (ethi->eth) > 1) ? 0 : 8),
+			  ((e_table_sort_info_get_can_group (ethi->sort_info)) ? 0 : 16) +
+			  128, info);
 }
 
 static void
@@ -1308,6 +1358,10 @@ ethi_change_sort_state (ETableHeaderItem *ethi, gdouble x)
 	int found = FALSE;
 	
 	col = e_table_header_get_column (ethi->eth, ethi_find_col_by_x (ethi, x));
+
+	if (col == NULL)
+		return;
+
 	model_col = col->col_idx;
 	
 	length = e_table_sort_info_grouping_get_count(ethi->sort_info);
@@ -1352,7 +1406,9 @@ ethi_change_sort_state (ETableHeaderItem *ethi, gdouble x)
 	}
 
 	if (!found) {
-		ETableSortColumn column = { model_col, 1 };
+		ETableSortColumn column;
+		column.column = model_col;
+		column.ascending = 1;
 		length = e_table_sort_info_sorting_get_count(ethi->sort_info);
 		if (length == 0)
 			length++;
@@ -1520,6 +1576,8 @@ ethi_class_init (GtkObjectClass *object_class)
 				 GTK_ARG_WRITABLE, ARG_SORT_INFO);
 	gtk_object_add_arg_type ("ETableHeaderItem::table", GTK_TYPE_OBJECT,
 				 GTK_ARG_WRITABLE, ARG_TABLE);
+	gtk_object_add_arg_type ("ETableHeaderItem::tree", E_TREE_TYPE,
+				 GTK_ARG_WRITABLE, ARG_TREE);
 
 	/*
 	 * Create our pixmaps for DnD
@@ -1536,12 +1594,12 @@ ethi_class_init (GtkObjectClass *object_class)
 	ethi_signals [BUTTON_PRESSED] =
 		gtk_signal_new ("button_pressed",
 				GTK_RUN_LAST,
-				object_class->type,
+				E_OBJECT_CLASS_TYPE (object_class),
 				GTK_SIGNAL_OFFSET (ETableHeaderItemClass, button_pressed),
 				gtk_marshal_NONE__POINTER,
 				GTK_TYPE_NONE, 1, GTK_TYPE_GDK_EVENT);
 		
-	gtk_object_class_add_signals (object_class, ethi_signals, LAST_SIGNAL);
+	E_OBJECT_CLASS_ADD_SIGNALS (object_class, ethi_signals, LAST_SIGNAL);
 }
 
 static void
@@ -1566,6 +1624,7 @@ ethi_init (GnomeCanvasItem *item)
 
 	ethi->group_indent_width = 0;
 	ethi->table = NULL;
+	ethi->tree = NULL;
 }
 
 GtkType
